import pkg_resources
from click import Choice, command, option

from gstui.gsclient import CachedClient
from gstui.ui_base import FzfUI


@command()
@option(
    "--interface", "-i", default="fzf", type=Choice(["fzf"]),
    help="UI interface to use"
)
@option("--clean", "-c", is_flag=True, help="Clean cache")
@option("--cache-path", "-p", default="~/.cache/gstui", help="Cache directory")
@option("--cache-all", "-a", is_flag=True, help="Cache all tree structure")
@option("--version", "-v", is_flag=True, help="Display version")
def main(interface, clean, cache_path, cache_all, version):
    storage_client = CachedClient()
    storage_client.cache_path = cache_path
    if version:
        print(pkg_resources.get_distribution("gstui").version)
        return
    if clean:
        storage_client.clear_cache()
        return
    if cache_all:
        storage_client.cache_all()
        return
    if interface == "fzf":
        ui = FzfUI()
        ui.mainloop(storage_client)
    storage_client.close()



from abc import abstractmethod
from pathlib import Path
from dataclasses import dataclass
from queue import LifoQueue
from typing import List, Optional

from pyfzf.pyfzf import FzfPrompt

from gstui.gsclient import CachedClient


@dataclass
class View:
    items: List[str]
    title: Optional[str]
    multi: bool = False


class UIBase:
    """Base class for stack of list view UIs"""

    def __init__(self):
        self.view_stack = LifoQueue()

    @abstractmethod
    def list_view(self, view: View) -> List[str]:
        """List view of the items"""
        pass

    @abstractmethod
    def mainloop(self, client: CachedClient):
        """Main loop"""
        pass

    def push(self, view: View):
        """Adds a list view to the stack"""
        values = self.list_view(view)
        self.view_stack.put(view)
        return values

    def pop(self) -> View:
        """Pops a list view from the stack"""
        return self.view_stack.get()


class FzfUI(UIBase):
    """Default UI"""

    def __init__(self, fzf: FzfPrompt = FzfPrompt()):
        self.fzf = fzf
        super().__init__()

    def list_view(self, view: View) -> List[str]:
        items = view.items
        title = view.title
        multi = view.multi
        return self.fzf.prompt(items,
                               f"{'--multi ' * multi}--prompt='{title}'")

    def mainloop(self, storage_client: CachedClient):
        buckets = storage_client.list_buckets()
        bucket_name = None
        while True:
            try:
                if bucket_name is None:
                    view = View(buckets, "Bucket: ")
                    bucket_name = self.push(view)
                    if not bucket_name:
                        break
                    bucket_name = bucket_name[0]
                else:
                    blobs = storage_client.list_blobs(bucket_name)
                    blob_names = self.push(View(blobs, "Blob: ", True))
                    if not blob_names:
                        bucket_name = None
                        self.pop()
                        self.pop()
                        continue
                    self.pop()
                    for blob_name in blob_names:
                        bucket = storage_client.bucket(bucket_name)
                        blob = bucket.get_blob(blob_name)
                        if blob:
                            storage_client.download(blob, Path(blob_name).name)
            except KeyboardInterrupt:
                return



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gstui']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'blessed>=1.19.1,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'diskcache>=5.4.0,<6.0.0',
 'google-cloud-storage>=2.5.0,<3.0.0',
 'pyfzf>=0.3.1,<0.4.0',
 'pytermgui>=7.2.0,<8.0.0',
 'tqdm>=4.64.1,<5.0.0']

entry_points = \
{'console_scripts': ['gstui = gstui.gstui:main', 'tests = scripts:test']}

setup_kwargs = {
    'name': 'gstui',
    'version': '0.1.4',
    'description': 'Text User interface for exploring, downloading and uploading files to Google Cloud Storage buckets',
    'long_description': "# Work in progress\n\n# GSTUI\n\nA Text User Interface for exploring Google Cloud Storage. Fast and cached.\n\n## Intallation\n\n```sh\npip install -U gstui\n```\n\nInstall [fzf](https://github.com/junegunn/fzf#installation)\n\n## Usage\n\nRun `gstui` or `gstui --help` to see more options.\n\nLoading buckets or the inital listing for the first time can take a long time to cache. You can create an initial cache of everything with: `gstui -a`.\n\nThe first picker is for selecting the bucket and the second is for selecting the blob to download.\n\n# Development\n\nBe free to submit a PR. Check the formatting with flake8 and for new features try to write tests.\n\n## Tests\n\n\n```sh\npoetry run tests\n```\n\nOr manually\n\n```sh\npoetry run pytest tests -n 4 -vvv\n```\n\n## TODO\n\n- [ ] Better thread management\n- [ ] Don't rely on `time.sleep` for cache tests\n- [ ] [urwid](https://github.com/urwid/urwid) UI\n\n# Related Projects\n\n* [gsutil](https://github.com/GoogleCloudPlatform/gsutil) A command line tool for interacting with cloud storage services. \n* [gcsfuse](https://github.com/GoogleCloudPlatform/gcsfuse) A user-space file system for interacting with Google Cloud Storage \n",
    'author': 'Matheus Fillipe',
    'author_email': 'matheusfillipeag@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import os
from pathlib import Path
from shutil import rmtree
from threading import Thread
from typing import List

from diskcache import Cache
from google.cloud import storage
from tqdm.std import tqdm


class GsClient(storage.Client):
    def download(self, *_):
        pass


def get_cache_path() -> str:
    """Get cache path in different OS's"""
    if os.name == "nt":
        path = os.getenv("LOCALAPPDATA")
        if path is None:
            path = os.getenv("APPDATA")
        if path is None:
            print("Cannot find cache path")
            exit(1)
            return
        path = Path(path) / "gstui" / "cache"
    else:
        path = Path.home() / ".cache" / "gstui"
    return str(path)


class ThreadedCachedClient:
    """
    Cached client. Spawns a thread with the google client.
    In case something is not cached, it will join the thread and block.
    """

    init_thread: Thread
    thread_pool: List[Thread]
    cache_path: str = get_cache_path()

    def spawn(self, cls, *args, **kwargs):
        # spawn init in thread
        self.init_thread = Thread(
            target=cls.__init__, args=(self, *args), kwargs=kwargs
        )
        self.init_thread.start()
        self.thread_pool = []

    def clear_cache(self):
        """Clean cache"""
        print("Cleaning cache database...")
        rmtree(Path(self.cache_path).expanduser(), ignore_errors=True)

    def _thread_pool_cleanup(self):
        """Cleanup thread pool"""
        for thread in self.thread_pool:
            if not thread.is_alive():
                self.thread_pool.remove(thread)

    def close(self):
        """Stop all threads"""
        self._thread_pool_cleanup()
        print("Stopping all threads...")
        if self.init_thread.is_alive():
            self.init_thread.join()
        for thread in self.thread_pool:
            thread.join()

    @classmethod
    def diskcache(cls, func):
        """Decorator to cache function results to disk"""

        def wrapper(self, *args, **kwargs):
            if not isinstance(self, ThreadedCachedClient):
                raise TypeError(
                    "Decorator only works with ThreadedCachedClient")
            cache = Cache(cls.cache_path)
            key = func.__name__ + ":" + str(args) + str(kwargs)
            result = cache.get(key)
            if result is None:
                # Ensure init thread is finished otherwise join it
                if self.init_thread.is_alive():
                    self.init_thread.join()
                result = func(self, *args, **kwargs)
                cache.set(key, result)
            elif not self.init_thread.is_alive():
                new_thread = Thread(target=func, args=(
                    self, *args), kwargs=kwargs)
                self.thread_pool.append(new_thread)
                new_thread.start()
                self._thread_pool_cleanup()

            return result

        return wrapper


class CachedClient(GsClient, ThreadedCachedClient):
    """Google cloud storage ThreadedCachedClient"""

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.spawn(GsClient, *args, **kwargs)

    def close(self):
        super().close()
        ThreadedCachedClient.close(self)

    @ThreadedCachedClient.diskcache
    def list_buckets(self, *args, **kwargs) -> List[str]:
        return [
            bucket.name for bucket in super().list_buckets(*args, **kwargs)
        ]

    @ThreadedCachedClient.diskcache
    def list_blobs(self, *args, **kwargs) -> List[str]:
        return [blob.name for blob in super().list_blobs(*args, **kwargs)]

    def cache_all(self):
        """Cache all tree structure"""
        print("Caching all tree structure. This might take a while...")
        buckets = self.list_buckets()
        for bucket in tqdm(buckets):
            self.list_blobs(bucket)

    # TODO part of this should be responsability of the UI instead
    def download(self, blob: storage.Blob, destination_file_name: str):
        if blob.size is not None:
            print(f"Downloading {blob.size/1024/1024:.2f} MB")
        with open(destination_file_name, "wb") as f:
            with tqdm.wrapattr(f, "write", total=blob.size) as file_obj:
                self.download_blob_to_file(blob, file_obj)
        print(f"Downloaded {destination_file_name}")
